@@ -15,14 +15,30 @@ BEGIN {
   *_HAVE_IS_UTF8 = defined &utf8::is_utf8 ? sub(){1} : sub(){0};
   *_HAVE_PERLSTRING = defined &B::perlstring ? sub(){1} : sub(){0};
   *_BAD_BACKSLASH_ESCAPE = _HAVE_PERLSTRING() && "$]" == 5.010_000 ? sub(){1} : sub(){0};
-  *_HAVE_HEX_FLOAT = "$]" >= 5.022 ? sub(){1} : sub(){0};
+  *_HAVE_HEX_FLOAT = !$ENV{SUB_QUOTE_NO_HEX_FLOAT} && "$]" >= 5.022 ? sub(){1} : sub(){0};
 
-  my $precision = length 1/9;
-  *_MAX_FLOAT_PRECISION = sub(){$precision};
+  # This may not be perfect, as we can't tell the format purely from the size
+  # but it should cover the common cases, and other formats are more likely to
+  # be less precise.
+  my $nvsize = 8 * length pack 'F', 0;
+  my $nvmantbits
+    = $nvsize == 16   ? 11
+    : $nvsize == 32   ? 24
+    : $nvsize == 64   ? 53
+    : $nvsize == 80   ? 64
+    : $nvsize == 128  ? 113
+    : $nvsize == 256  ? 237
+                      : 237 # unknown float format
+    ;
+  my $precision = int( log(2)/log(10)*$nvmantbits );
+
+  *_NVSIZE = sub(){$nvsize};
+  *_NVMANTBITS = sub(){$nvmantbits};
+  *_FLOAT_PRECISION = sub(){$precision};
 }
 
-our $VERSION = '2.006_004';
-$VERSION = eval $VERSION;
+our $VERSION = '2.006_005';
+$VERSION =~ tr/_//d;
 
 our @EXPORT = qw(quote_sub unquote_sub quoted_from_sub qsub);
 our @EXPORT_OK = qw(quotify capture_unroll inlinify sanitize_identifier);
@@ -54,21 +70,44 @@ sub quotify {
     && 0 + $value eq $value
   ) ? (
     $value != $value ? (
-      $value eq -(9**9**9*0)
-        ? '(-(9**9**9*0))' # -nan
-        : '(9**9**9*0)'    # nan
+      $value eq (9**9**9*0)
+        ? '(9**9**9*0)'    # nan
+        : '(-(9**9**9*0))' # -nan
     )
-    : $value == 9**9**9 ? '(9**9**9)'      # inf
+    : $value == 9**9**9  ? '(9**9**9)'     # inf
     : $value == -9**9**9 ? '(-9**9**9)'    # -inf
+    : $value == 0 ? (
+      sprintf('%g', $value) eq '-0' ? '-0.0' : '0',
+    )
+    : $value !~ /[e.]/i ? (
+      $value > 0 ? (sprintf '%u', $value)
+                 : (sprintf '%d', $value)
+    )
     : do {
       my $float = $value;
-      for my $precision (0 .. _MAX_FLOAT_PRECISION) {
-        $float = sprintf '%.'.$precision.'g', $value;
-        last
-          if $float == $value;
-
-        if (_HAVE_HEX_FLOAT and $precision == _MAX_FLOAT_PRECISION) {
-          $float = sprintf "%a", $value;
+      my $max_factor = int( log( abs($value) ) / log(2) ) - _NVMANTBITS;
+      my $ex_sign = $max_factor > 0 ? 1 : -1;
+      FACTOR: for my $ex (0 .. abs($max_factor)) {
+        my $num = $value / 2**($ex_sign * $ex);
+        for my $precision (_FLOAT_PRECISION .. _FLOAT_PRECISION+2) {
+          my $formatted = sprintf '%.'.$precision.'g', $num;
+          if ($formatted == $num) {
+            $float = $formatted;
+            if ($ex) {
+              $float
+                .= ($ex_sign == 1 ? '*' : '/')
+                . (
+                  $ex > _NVMANTBITS
+                    ? "2**$ex"
+                    : sprintf('%u', 2**$ex)
+                );
+            }
+            last FACTOR;
+          }
+        }
+        if (_HAVE_HEX_FLOAT) {
+          $float = sprintf '%a', $value;
+          last FACTOR;
         }
       }
       "$float";
